"""
The MIT License (MIT)

Copyright (c) 2023-present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

from typing import TypedDict

from .shared import *

__all__ = ("ConfigFile",)


class ConfigFile(TypedDict):
    TimesLanded: IntValue
    PlayerXPNum: IntValue
    PlayerLevel: IntValue
    FinishedShockMinigame: IntValue
    HostSettings_Public: BoolValue
    HostSettings_Name: StringValue
    LastVerPlayed: IntValue
    SpiderSafeMode: BoolValue
    InvertYAxis: BoolValue
    ScreenMode: IntValue
    FPSCap: IntValue
    Bindings: StringValue
    CurrentMic: StringValue
    PushToTalk: BoolValue
    MicEnabled: BoolValue
    LookSens: IntValue
    MasterVolume: FloatValue
    Gamma: FloatValue
    StartInOnlineMode: BoolValue
    PlayerFinishedSetup: BoolValue
    SelectedFile: IntValue
    LC_MoveObjectsTip: BoolValue
    LC_StorageTip: BoolValue
    LC_EclipseTip: BoolValue
    PlayedDungeonEntrance0: BoolValue
    PlayedDungeonEntrance1: BoolValue
    HasUsedTerminal: BoolValue
    LCTip_SellScrap: BoolValue
    LCTip_SecureDoors: BoolValue
    LCTip_UseManual: BoolValue
    LC_IntroTip1: BoolValue



"""
The MIT License (MIT)

Copyright (c) 2023-present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import random
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from .crypt import decrypt, encrypt
from .enums import BestiaryEntry, ExtraUnlock, Item, Moon, Scrap, ShipUnlock
from .item import GrabbableScrap
from .utils import MISSING, SaveValue, _to_json, resolve_save_path  # type: ignore[reportPrivateUsage] we allow this here.
from .vector import Vector

if TYPE_CHECKING:
    from os import PathLike
    from types import TracebackType

    from typing_extensions import Self

    from .types_.challenge_file import ChallengeFile as ChallengeFileType
    from .types_.config_file import ConfigFile as ConfigFileType
    from .types_.save_file import (
        SaveFile as SaveFileType,
    )
    from .types_.shared import *

SaveT = TypeVar("SaveT", "SaveFileType", "ConfigFileType", "ChallengeFileType")

TEMP_FILE = Path("./_previously_decrypted_file.json")
TIPS = [
    "LC_MoveObjectsTip",
    "LC_StorageTip",
    "LC_LightningTip",
    "LCTip_SecureDoors",
    "LC_EclipseTip",
    "LCTip_SellScrap",
    "LCTip_UseManual",
    "LC_IntroTip1",
]

__all__ = (
    "SaveFile",
    "ConfigFile",
    "ChallengeFile",
)


class _BaseSaveFile(Generic[SaveT]):
    _inner_data: SaveT
    _file_type: str
    _extra_data: dict[str, Any]
    _written: bool
    _skip_parsing: bool

    __slots__ = (
        "_inner_data",
        "_file_type",
        "_extra_data",
        "_written",
        "_skip_parsing",
        "_raw_data",
    )

    def __init__(self, data: bytes, /) -> None:
        self._skip_parsing = False
        self._written = False

        self._raw_data: bytes = data
        self._parse_file()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        if not self._written and not exc_type:
            self.write()

    @classmethod
    def from_path(cls, path: Path | PathLike[Any] | str) -> Self:
        if not isinstance(path, Path):
            path = Path(path)

        if not path.exists():
            raise FileNotFoundError("The passed file is not found.")

        with path.open("rb") as fp:
            return cls(fp.read())

    def _parse_file(self) -> None:
        if self._skip_parsing:
            return

        data = decrypt(data=self._raw_data)

        self._validate_contents(data)

        self._inner_data = data

    def _upsert_value(self, key_name: str, value: Any) -> None:
        if value is MISSING:
            return  # If the value is the sentinel type, do nothing and move onto the next

        if isinstance(value, int):
            _type = "int"
        elif isinstance(value, list):
            if value and isinstance(value[0], int):
                _type = "System.Int32[],mscorlib"
            elif value and isinstance(value[0], dict):
                _type = "UnityEngine.Vector3[],UnityEngine.CoreModule"
            else:
                raise ValueError("Unexpected or unknown array type passed for `value`")
        elif isinstance(value, bool):
            _type = "bool"
        else:
            raise ValueError("Unexpected type passed for `value`: %r (%s)", value, type(value))

        try:
            self._inner_data[key_name]["value"] = value
        except KeyError:
            self._inner_data[key_name] = {"__type": _type, "value": value}

    def _dump(self, path: Path, /) -> None:
        decrypted_result = _to_json(self._inner_data)

        encoded = decrypted_result.encode()

        with TEMP_FILE.open("wb") as fp:
            fp.write(encoded)

        encrypted_result = encrypt(data=encoded)

        with path.open("wb") as fp:
            fp.write(encrypted_result)

    def write(self, *, path: Path | None = None) -> None:
        for key, value in self._extra_data.items():
            self._upsert_value(key, value)

        if path:
            self._dump(path)

        self._written = True

    def _validate_contents(self, data: SaveT, /) -> None:
        raise NotImplementedError


class SaveFile(_BaseSaveFile["SaveFileType"]):
    """
    A Lethal Company save file.

    Parameters
    -----------
    data: :class:`bytes`
        The data read from the save file.
    """

    # late init variable types
    _extra_data: dict[str, Any]

    _credits: int
    _current_planet_id: int
    _deadline: float
    _deaths: int
    _elapsed_days: int
    _quotas_met: int
    _quota_threshold: int
    _seed: int
    _steps_taken: int

    __slots__ = (
        "_inner_data",
        "_extra_data",
        "_credits",
        "_current_planet_id",
        "_current_quota_progress",
        "_deadline",
        "_deaths",
        "_elapsed_days",
        "_quotas_met",
        "_quota_threshold",
        "_seed",
        "_steps_taken",
        # these values need a richer interface
        "_enemy_scans",
        "_ship_item_save_data",
        "_unlocked_ship_objects",
        "_scrap",
        "__ship_grabbable_items",
        "__ship_grabbable_item_positions",
        "__ship_scrap",
    )

    @classmethod
    def resolve_from_file(cls, save_file_number: SaveValue, /) -> SaveFile:
        """
        A class method to find, open and return a save file from the default save file location (Windows) with the given file number.

        Parameters
        -----------
        save_file_number: Literal[``1``, ``2``, ``3``]
            The numbered save to open.
        """
        path = resolve_save_path(save_file_number)

        return cls.from_path(path)

    def _validate_contents(self, data: SaveFileType, /) -> None:
        _required_keys = (
            "GroupCredits",
            "DeadlineTime",
            "Stats_StepsTaken",
            "Stats_DaysSpent",
            "ProfitQuota",
            "CurrentPlanetID",
        )
        if any(key not in data for key in _required_keys):
            raise ValueError("This doesn't appear to be a valid Lethal Company save file!")

    def _generate_seed(self, *, max: int = 99999999, min: int = 10000000) -> int:
        return random.randint(min, max)

    def _parse_file(self) -> None:
        super()._parse_file()

        self._credits = self._inner_data["GroupCredits"]["value"]
        self._current_planet_id = self._inner_data["CurrentPlanetID"]["value"]
        self._current_quota_progress = self._inner_data["QuotaFulfilled"]["value"]
        self._deadline = self._inner_data["DeadlineTime"]["value"]
        self._deaths = self._inner_data["Stats_Deaths"]["value"]
        self._elapsed_days = self._inner_data["Stats_DaysSpent"]["value"]
        self._quotas_met = self._inner_data["QuotasPassed"]["value"]
        self._quota_threshold = self._inner_data["ProfitQuota"]["value"]
        self._seed = self._inner_data["RandomSeed"]["value"]
        self._steps_taken = self._inner_data["Stats_StepsTaken"]["value"]

        # TODO: richer interface here.
        self._enemy_scans = self._inner_data.get("EnemyScans", {"__type": "System.Int32[],mscorlib", "value": []})
        self._ship_item_save_data = self._inner_data.get(
            "shipItemSaveData", {"__type": "System.Int32[],mscorlib", "value": []}
        )
        self._unlocked_ship_objects = self._inner_data.get(
            "UnlockedShipObjects", {"__type": "System.Int32[],mscorlib", "value": []}
        )

        self.__ship_grabbable_items = self._inner_data.get(
            "shipGrabbableItemIDs", {"__type": "System.Int32[],mscorlib", "value": []}
        )
        self.__ship_grabbable_item_positions = self._inner_data.get(
            "shipGrabbableItemPos", {"__type": "UnityEngine.Vector3[],UnityEngine.CoreModule", "value": []}
        )
        self.__ship_scrap = self._inner_data.get("shipScrapValues", {"__type": "System.Int32[],mscorlib", "value": []})
        self._parse_scrap_mapping()

        # this key is mostly laziness for now
        # we'll serialise anything in here into the final payload
        # for now this will just be how we add the UnlockedStored_X keys
        self._extra_data = {}

    def _parse_scrap_mapping(self) -> None:
        # shipGrabbableItems contains all touchable items on the ship, including tools which have no value
        # shipScrapValues are an array of values assigned to each piece of scrap
        # it works because GrabbableItems[1]: ScrapValues[1], each index aligns and that's how the values are assigned, like a zip
        # once the scrapvalues runs out of elements, the rest of the items are treated as no value, like tools
        self._scrap: list[GrabbableScrap] = []

        for item, value, pos in zip(
            self.__ship_grabbable_items["value"], self.__ship_scrap["value"], self.__ship_grabbable_item_positions["value"]
        ):
            self._scrap.append(GrabbableScrap(item, value, pos))

        self.__ship_grabbable_items["value"] = self.__ship_grabbable_items["value"][len(self._scrap) :]
        self.__ship_grabbable_item_positions["value"] = self.__ship_grabbable_item_positions["value"][len(self._scrap) :]

    @property
    def credits(self) -> int:
        """
        Get the current credits value within the save.

        Returns
        --------
        :class:`int`
        """
        return self._credits

    @property
    def current_moon(self) -> Moon:
        """
        Get the current planet within the save.

        Returns
        --------
        :class:`~great_asset.Moon`
        """
        return Moon(self._current_planet_id)

    @property
    def steps_taken(self) -> int:
        """
        Get the current amount of steps taken within the save file.

        Returns
        --------
        :class:`int`
        """
        return self._steps_taken

    @property
    def deaths(self) -> int:
        """
        Get the current amount of deaths within the save file.

        Returns
        --------
        :class:`int`
        """
        return self._deaths

    @property
    def elapsed_days(self) -> int:
        """
        Get the current amount of elapsed days within the save file.

        Returns
        --------
        :class:`int`
        """
        return self._elapsed_days

    @property
    def deadline(self) -> int:
        """
        Get the current deadline in time as days within the save file.


        .. warning:
            I round this value using :meth:`round` so it may not be fully accurate.
            If you need the raw value, try :property:`raw_deadline`.

        Returns
        --------
        :class:`int`
        """
        return round(self._deadline / 1080)

    @property
    def raw_deadline(self) -> float:
        """
        Get the current deadline in time's raw value within the save file.

        Returns
        --------
        :class:`float`
        """
        return self._deadline

    @property
    def profit_quota(self) -> int:
        """
        Get the current profit quota (value to reach) within the save file.

        Returns
        --------
        :class:`int`
        """
        return self._quota_threshold

    @property
    def quotas_passed(self) -> int:
        """
        Get the current total quotas met from within the save file.

        Returns
        --------
        :class:`int`
        """
        return self._quotas_met

    @property
    def current_quota_progress(self) -> int:
        """
        Get the current quota progress from within the save file.

        Returns
        --------
        :class:`int`
        """
        return self._current_quota_progress

    @property
    def current_seed(self) -> int:
        """
        Get the current seed from within the save file.

        Returns
        --------
        :class:`int`
        """
        return self._seed

    def update_credits(self, new_credits: int, /) -> None:
        """Update the credits value within the save file.

        Parameters
        -----------
        new_credits: :class:`int`
            The new credits value.
        """
        self._upsert_value("GroupCredits", new_credits)
        self._credits = new_credits

    def update_current_moon(self, moon: Moon, /) -> None:
        """
        Update the current planet within the save file.

        Parameters
        -----------
        planet: :class:`~great_asset.Moon`
            The planet to update to.
        """
        self._upsert_value("CurrentPlanetID", moon.value)
        self._current_planet_id = moon.value

    def update_steps_taken(self, new_steps: int, /) -> None:
        """
        Update the current amount of steps taken within the save file.

        Parameters
        -----------
        new_steps: :class:`int`
            The amount of steps to have taken.
        """
        self._upsert_value("Stats_StepsTaken", new_steps)
        self._steps_taken = new_steps

    def update_deaths(self, new_deaths: int, /) -> None:
        """
        Update the current deaths within the save file.

        Parameters
        -----------
        new_deaths: :class:`int`
            The new value for total deaths.
        """
        self._upsert_value("Stats_Deaths", new_deaths)
        self._deaths = new_deaths

    def update_elapsed_days(self, new_elapsed_days: int, /) -> None:
        """
        Update the elapsed days count within the save file.

        Parameters
        -----------
        new_elapsed_days: :class:`int`
            The elapsed days value.
        """
        self._upsert_value("Stats_DaysSpent", new_elapsed_days)
        self._elapsed_days = new_elapsed_days

    def update_deadline(self, new_deadline: int | float, /) -> None:
        """
        Update the deadline time within the save file.

        Parameters
        -----------
        new_deadline: :class:`int` | :class:`float`
            New deadline/time remaining in days or minutes.
        """
        if not isinstance(new_deadline, float):
            new_deadline = new_deadline * 1080

        self._upsert_value("DeadlineTime", new_deadline)
        self._deadline = new_deadline

    def update_profit_quota(self, new_profit_quota: int, /) -> None:
        """
        Update the target quota value within the save file.

        Parameters
        -----------
        new_profit_quota: :class:`int`
            The profit quota to set.
        """
        self._upsert_value("ProfitQuota", new_profit_quota)
        self._quota_threshold = new_profit_quota

    def update_quotas_met(self, new_quotas_met: int, /) -> None:
        """
        Updates the quotas met within the save file.

        Parameters
        -----------
        new_quotas_met: :class:`int`
            The quotas met to set.
        """
        self._upsert_value("QuotasPassed", new_quotas_met)
        self._quotas_met = new_quotas_met

    def update_current_quota_progress(self, new_quota_progress: int, /) -> None:
        """
        Updates the current quota progress within the save file.

        Parameters
        -----------
        new_quota_progress: :class:`int`
            The quota progress to set.
        """
        self._upsert_value("QuotaFulfilled", new_quota_progress)
        self._current_quota_progress = new_quota_progress

    def update_current_seed(self, new_seed: int | None = None, /) -> None:
        """
        Updates the current seed within the save file.

        Parameters
        -----------
        new_seed: :class:`int` | :class:`None`
            The new seed to update to. If ``None`` is passed or no value is given a random one will be generated.
        """
        seed = new_seed or self._generate_seed()

        self._upsert_value("RandomSeed", seed)
        self._seed = seed

    def unlock_ship_upgrades(self, *items: ShipUnlock) -> None:
        """
        Unlock upgrades for the ship.

        Parameters
        -----------
        *items: :class:`~great_asset.ShipUnlock`
            The items to unlock in the ship.


        .. note::
            The items unlocked will be added into storage, please access them from the terminal to place them.
        """
        for item in items:
            self._unlocked_ship_objects["value"].append(item.serialised_value)
            self._extra_data[f"ShipUnlockStored_{item.serialised_name}"] = True

    def unlock_all_ship_upgrades(self) -> None:
        """
        Unlocks all possible ship upgrades.
        """
        return self.unlock_ship_upgrades(*ShipUnlock.all())

    def remove_ship_upgrades(self, *items: ShipUnlock) -> None:
        """
        Remove upgrades from the ship.

        Parameters
        -----------
        *items: :class:`~great_asset.ShipUnlock`
            The items to remove from the ship.
        """
        for item in items:
            try:
                self._unlocked_ship_objects["value"].remove(item.serialised_value)
            except ValueError:
                pass
            self._extra_data[f"ShipUnlockStored_{item.serialised_name}"] = False

    def unlock_extras(self, *items: ExtraUnlock) -> None:
        """
        Unlock other items within the ship or for the player(s).

        Parameters
        -----------
        *items: :class:`~great_asset.ExtraUnlock`
            The items to unlock.
        """
        self._unlocked_ship_objects["value"] = list({i.value for i in items})

    def unlock_all_ship_extras(self) -> None:
        """
        Unlock all possible ship extras.
        """
        return self.unlock_extras(*ExtraUnlock.all())

    def unlock_bestiary_entries(self, *entries: BestiaryEntry) -> None:
        """
        Unlock bestiary entries on the ship terminal.

        Parameters
        -----------
        *entries: :class:`~great_asset.BestiaryEntry`
            The entries to unlock.
        """
        self._enemy_scans["value"] = list({e.value for e in entries})

    def unlock_all_bestiary_entries(self) -> None:
        """
        Unlock all possible bestiary entries.
        """
        return self.unlock_bestiary_entries(*BestiaryEntry.all())

    def spawn_items(self, *items: tuple[Item | Scrap, Vector | None], value_min: int = 30, value_max: int = 90) -> None:
        """
        Spawn items within the world or ship.

        Parameters
        -----------
        *items: tuple[:class:`~great_asset.Item` | :class:`~great_asset.Scrap`, :class:`~great_asset.Vector` | :class:`None`]
            A series of tuples with the item and it's position to spawn in.
            Using ``None`` as the second value will spawn at a default area near the door of the ship internally.
        """
        cupboard_position = self._inner_data.get("ShipUnlockPos_Cupboard")

        for item, position in items:
            vec = position or Vector.in_cupboard(cupboard_position=cupboard_position)
            if isinstance(item, Scrap):
                value = random.randint(value_min, value_max)
                self._scrap.append(GrabbableScrap(item.value, value, vec.serialise()))
                continue

            self.__ship_grabbable_items["value"].append(item.value)
            self.__ship_grabbable_item_positions["value"].append(vec.serialise())

    def get_current_items(self) -> list[Item]:
        """
        Get the current items that exist within the ship in the save file.

        Returns
        --------
        list[:class:`~great_asset.Item`]
        """
        return [Item(item) for item in self.__ship_grabbable_items["value"]]

    def get_current_scrap(self) -> list[Scrap]:
        """
        Get the current scrap that exist within the ship in the save file.

        Returns
        --------
        list[:class:`~great_asset.Scrap`]
        """
        return [Scrap(item.id) for item in self._scrap]

    def write(self, *, path: Path | None = None) -> None:
        """
        A function to write the save file data to the internal data structure.

        Parameters
        -----------
        dump_to_file: :class:`bool`
            Whether to overwrite the passed file with the changes. Defaults to ``True``.
        overwrite: :class:`bool`
            Whether to overwrite the existing save file with the changes. Defaults to ``True``.
            Irrelevant if ``dump_to_file`` is ``False``.
            Creates a file with the same name with the file extension of ``".over"``.
        """
        # manually handle the more complex types:
        self._upsert_value("UnlockedShipObjects", list(set(self._unlocked_ship_objects["value"])))
        self._inner_data["EnemyScans"] = self._enemy_scans
        self._inner_data["shipItemSaveData"] = self._ship_item_save_data

        for item in self._scrap:
            self.__ship_scrap["value"].insert(0, item.value)
            self.__ship_grabbable_items["value"].insert(0, item.id)
            self.__ship_grabbable_item_positions["value"].insert(0, item.pos)

        self._inner_data["shipScrapValues"] = self.__ship_scrap
        self._inner_data["shipGrabbableItemIDs"] = self.__ship_grabbable_items
        self._inner_data["shipGrabbableItemPos"] = self.__ship_grabbable_item_positions

        super().write(path=path)


class ConfigFile(_BaseSaveFile["ConfigFileType"]):
    _extra_data: dict[str, Any]

    # late init types
    arachnophobia_mode: bool
    y_axis_inverted: bool
    screen_mode: int
    fps_mode: int
    current_mic: str
    push_to_talk: bool
    mic_enabled: bool
    mouse_sensitivity: int
    master_volume: float
    gamma: float
    finished_setup: bool
    start_online: bool

    _player_xp: IntValue
    _player_level: IntValue
    _times_landed: IntValue
    _is_host_public: BoolValue
    _host_name: StringValue
    _played_extrance_0: BoolValue
    _played_entrance_1: BoolValue
    _tips: dict[str, BoolValue]

    def _validate_contents(self, data: ConfigFileType, /) -> None:
        _required_keys = ("SelectedFile", "FPSCap", "Gamma", "LookSens", "ScreenMode")
        if any(key not in data for key in _required_keys):
            raise ValueError("This doesn't appear to be a valid Lethal Company config file!")

    def _parse_file(self) -> None:
        super()._parse_file()

        self.arachnophobia_mode = self._inner_data["SpiderSafeMode"]["value"]
        self.y_axis_inverted = self._inner_data["InvertYAxis"]["value"]
        self.screen_mode = self._inner_data["ScreenMode"]["value"]
        self.fps_mode = self._inner_data["FPSCap"]["value"]
        self.current_mic = self._inner_data["CurrentMic"]["value"]
        self.push_to_talk = self._inner_data["PushToTalk"]["value"]
        self.mic_enabled = self._inner_data["MicEnabled"]["value"]
        self.mouse_sensitivity = self._inner_data["LookSens"]["value"]
        self.master_volume = self._inner_data["MasterVolume"]["value"]
        self.gamma = self._inner_data["Gamma"]["value"]
        self.finished_setup = self._inner_data["PlayerFinishedSetup"]["value"]
        self.start_online = self._inner_data["StartInOnlineMode"]["value"]

        self._player_xp = self._inner_data.get("PlayerXPNum", MISSING)
        self._player_level = self._inner_data.get("PlayerLevel", MISSING)
        self._times_landed = self._inner_data.get("TimesLanded", MISSING)
        self._is_host_public = self._inner_data.get("HostSettings_Public", MISSING)
        self._host_name = self._inner_data.get("HostSettings_Name", MISSING)
        self._played_entrance_0 = self._inner_data.get("PlayedDungeonEntrance0", MISSING)
        self._played_entrance_1 = self._inner_data.get("PlayedDungeonEntrance1", MISSING)
        self._tips = {tip: self._inner_data.get(tip, MISSING) for tip in TIPS}
        self._extra_data = {}

    @property
    def player_xp(self) -> int:
        if self._player_xp is not MISSING:
            return self._player_xp["value"]
        return 0

    @property
    def player_level(self) -> int:
        if self._player_xp is not MISSING:
            return self._player_level["value"]
        return 0

    @property
    def times_landed(self) -> int:
        if self._times_landed is not MISSING:
            return self._times_landed["value"]
        return 0

    @property
    def is_host_public(self) -> bool:
        if self._is_host_public is not MISSING:
            return self._is_host_public["value"]
        return False

    @property
    def host_name(self) -> str:
        if self._host_name is not MISSING:
            return self._host_name["value"]
        return "No name set"

    @property
    def played_entrances(self) -> list[int]:
        return [val["value"] for val in [self._played_entrance_0, self._played_entrance_1] if val is not MISSING]

    @property
    def tips(self) -> dict[str, bool]:
        return {tip: value["value"] for tip, value in self._tips.items() if value is not MISSING}

    def set_player_xp(self, value: int) -> None:
        if 0 <= value <= 999:
            self._player_xp = {"__type": "int", "value": value}

    def set_player_level(self, value: int) -> None:
        if value > 0:
            self._player_level = {"__type": "int", "value": (value - 1) % 5}

    def set_host_public(self, value: bool) -> None:
        self._is_host_public = {"__type": "bool", "value": value}

    def set_host_name(self, value: str) -> None:
        self._host_name = {"__type": "string", "value": value[40:]}

    def set_arachnophobia_mode(self, value: bool) -> None:
        self.arachnophobia_mode = value

    def set_inverted_y_axis(self, value: bool) -> None:
        self.y_axis_inverted = value

    def set_screen_mode(self, value: int) -> None:
        self.screen_mode = value

    def set_fps_mode(self, value: int) -> None:
        self.fps_mode = value

    def set_push_to_talk(self, value: bool) -> None:
        self.push_to_talk = value

    def set_mic_enabled(self, value: bool) -> None:
        self.mic_enabled = value

    def set_mouse_sensitivity(self, value: int) -> None:
        self.mouse_sensitivity = value

    def set_master_volume(self, value: float) -> None:
        if 0 <= value <= 1:
            self.master_volume = value

    def set_gamma(self, value: float) -> None:
        if 0 <= value <= 1:
            self.gamma = value

    def write(self, *, path: Path | None = None) -> None:
        self._upsert_value("SpiderSafeMode", self.arachnophobia_mode)
        self._upsert_value("InvertYAxis", self.y_axis_inverted)
        self._upsert_value("ScreenMode", self.screen_mode)
        self._upsert_value("FPSCap", self.fps_mode)
        self._upsert_value("CurrentMic", self.current_mic)
        self._upsert_value("PushToTalk", self.push_to_talk)
        self._upsert_value("MicEnabled", self.mic_enabled)
        self._upsert_value("LookSens", self.mouse_sensitivity)
        self._upsert_value("MasterVolume", self.master_volume)
        self._upsert_value("Gamma", self.gamma)
        self._upsert_value("PlayerFinishedSetup", self.finished_setup)
        self._upsert_value("StartInOnlineMode", self.start_online)
        self._upsert_value("PlayerXPNum", self._player_xp)
        self._upsert_value("PlayerLevel", self._player_level)
        self._upsert_value("TimesLanded", self._times_landed)
        self._upsert_value("HostSettings_Public", self._is_host_public)
        self._upsert_value("HostSettings_Name", self._host_name)
        self._upsert_value("PlayedDungeonEntrance0", self._played_entrance_0)
        self._upsert_value("PlayedDungeonEntrance1", self._played_entrance_1)

        for idx, tip in enumerate(self._tips):
            self._upsert_value(TIPS[idx], tip)

        super().write(path=path)


class ChallengeFile(_BaseSaveFile["ChallengeFileType"]):
    _profit_earned: int
    _finished_challenge: bool
    _submitted_score: bool
    _file_game_version: int
    _challenge_week_number: int
    _set_challenge_file_money: bool

    __slots__ = (
        "_profit_earned",
        "_finished_challenge",
        "_submitted_score",
        "_file_game_version",
        "_challenge_week_number",
        "_set_challenge_file_money",
    )

    def _validate_contents(self, data: ChallengeFileType) -> None:
        _required_keys = ("ProfitEarned", "FinishedChallenge", "SubmittedScore")
        if any(key not in data for key in _required_keys):
            raise ValueError("This doesn't appear to be a valid Challenge file.")

    def _parse_file(self) -> None:
        super()._parse_file()

        self._profit_earned = self._inner_data["ProfitEarned"]["value"]
        self._finished_challenge = self._inner_data["FinishedChallenge"]["value"]
        self._submitted_score = self._inner_data["SubmittedScore"]["value"]
        self._file_game_version = self._inner_data.get("FileGameVers", {"__type": "int", "value": 48})["value"]
        self._challenge_week_number = self._inner_data.get("FileGameVers", {"__type": "int", "value": 4})["value"]
        self._set_challenge_file_money = self._inner_data.get("SetChallengeFileMoney", {"__type": "bool", "value": False})[
            "value"
        ]

        self._extra_data = {}

    @property
    def profit_earned(self) -> int:
        return self._profit_earned

    @property
    def finished_challenge(self) -> bool:
        return self._finished_challenge

    @property
    def submitted_score(self) -> int:
        return self._submitted_score

    @property
    def file_game_version(self) -> int:
        return self._file_game_version

    @property
    def challenge_week_number(self) -> int:
        return self._challenge_week_number

    @property
    def set_challenge_file_money(self) -> bool:
        return self._set_challenge_file_money



"""
The MIT License (MIT)

Copyright (c) 2023-present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import json
import os
import pathlib
import platform
from typing import Any, Literal

__all__ = (
    "_to_json",
    "_from_json",
    "MISSING",
    "SaveValue",
    "resolve_save_path",
)

SaveValue = Literal[1, 2, 3, "1", "2", "3"]

try:
    import orjson
except ModuleNotFoundError:

    def _to_json(obj: Any, /) -> str:
        """A quick method that dumps a Python type to JSON object."""
        return json.dumps(obj, separators=(",", ":"), ensure_ascii=True, indent=2, sort_keys=True)

    _from_json = json.loads  # type: ignore # the overloads are too much for our use case.
else:

    def _to_json(obj: Any, /) -> str:
        """A quick method that dumps a Python type to JSON object."""
        return orjson.dumps(obj, option=orjson.OPT_INDENT_2 | orjson.OPT_SORT_KEYS).decode("utf-8")

    _from_json = orjson.loads  # type: ignore # this is guarded in an if.


class _MissingSentinel:
    __slots__ = ()

    def __eq__(self, other: object) -> bool:
        return False

    def __bool__(self) -> bool:
        return False

    def __hash__(self) -> int:
        return 0

    def __repr__(self) -> str:
        return "..."


MISSING: Any = _MissingSentinel()


def resolve_save_path(save_number: SaveValue, /) -> pathlib.Path:
    if platform.system() != "Windows":
        raise NotImplementedError("Currently we don't support non-Windows yet.")

    user_profile = os.getenv("USERPROFILE")
    if not user_profile:
        raise RuntimeError("This shouldn't happen or your windows profile is messed up.")

    save_data_path = pathlib.Path(user_profile) / "AppData" / "LocalLow" / "ZeekerssRBLX" / "Lethal Company"

    save_file_name = f"LCSaveFile{save_number}"

    return save_data_path / save_file_name



"""
The MIT License (MIT)

Copyright (c) 2023-present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

from Crypto import Random
from Crypto.Cipher import AES
from Crypto.Protocol.KDF import PBKDF2
from Crypto.Util.Padding import pad, unpad

from . import CRYPTO_PASSWORD
from .utils import _from_json  # type: ignore # allowing this private usage

if TYPE_CHECKING:
    from os import PathLike

__all__ = (
    "encrypt",
    "decrypt",
)


def encrypt(*, path: str | PathLike[str] | Path | None = None, data: bytes | None = None) -> bytes:
    if not path and not data:
        raise ValueError("Either `path` or `data` must be provided.")

    if path:
        if not isinstance(path, Path):
            path = Path(path)

        with path.open("rb") as fp:
            data_to_encrypt = fp.read()
    else:
        data_to_encrypt = data
        assert data_to_encrypt  # guarded earlier

    # Generate a random IV (Initialization Vector)
    init_vector = Random.new().read(16)

    # Derive the key using PBKDF2 with SHA1 hash algorithm
    key = PBKDF2(CRYPTO_PASSWORD, init_vector, dkLen=16, count=100)

    # Create AES cipher object
    cipher = AES.new(key, AES.MODE_CBC, init_vector)  # type: ignore # the upstream types aren't great

    # Pad the data with PKCS7 before encryption
    padded_data = pad(data_to_encrypt, AES.block_size, style="pkcs7")

    # Encrypt the data
    encrypted_data = init_vector + cipher.encrypt(padded_data)

    return encrypted_data


def decrypt(
    *, path: str | PathLike[str] | Path | None = None, data: bytes | None = None
) -> Any:  # it returns the type of file we decrypt but alas
    if not path and not data:
        raise ValueError("Either `path` or `data` must be provided.")

    if path:
        if not isinstance(path, Path):
            path = Path(path)

        with path.open("rb") as fp:
            read_data = fp.read()
    else:
        read_data = data
        assert read_data  # guarded earlier

    # The initialisation vector is the first 16 bytes of the save file.
    init_vector = read_data[:16]
    # then we take the proceeding N bytes as the data
    _to_decrypt = read_data[16:]

    # create the decryption key from the provided data
    decryption_key = PBKDF2(CRYPTO_PASSWORD, init_vector, dkLen=16, count=100)

    # with the key we create the needed cipher
    cipher = AES.new(decryption_key, AES.MODE_CBC, init_vector)  # type: ignore # the upstream types aren't great

    # and now we decrypt the data
    decrypted_data = unpad(cipher.decrypt(_to_decrypt), AES.block_size, style="pkcs7")

    # and it's always UTF-8
    resolved_data = decrypted_data.decode("utf-8")

    return _from_json(resolved_data)



"""
The MIT License (MIT)

Copyright (c) 2023-present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

from enum import Enum

__all__ = (
    "ShipUnlock",
    "ExtraUnlock",
    "BestiaryEntry",
    "Item",
    "Scrap",
    "Moon",
)


class ShipUnlock(Enum):
    cozy_lights = 4, "Cozy lights"
    teleporter = 5, "Teleporter"
    television = 6, "Television"
    tv = 6, "Television"
    cupboard = 7, "Cupboard"
    file_cabinet = 8, "File Cabinet"
    toilet = 9, "Toilet"
    shower = 10, "Shower"
    light_switch = 11, "Light switch"
    record_player = 12, "Record player"
    table = 13, "Table"
    romantic_table = 14, "Romantic table"
    bunkbeds = 15, "Bunkbeds"
    terminal = 16, "Terminal"
    signal_translator = 17, "Signal translator"
    signal_transmitter = 17, "Signal translator"
    loud_horn = 18, "Loud horn"
    inverse_teleporter = 19, "Inverse Teleporter"
    jack_o_lantern = 20, "JackOLantern"
    welcome_mat = 21, "Welcome mat"
    goldfish = 22, "Goldfish"
    plushie_pajama_man = 23, "Plushie pajama man"
    plushie_pyjama_man = 23, "Plushie pajama man"

    def __init__(self, value: int, serialised_name: str) -> None:
        self._serialised_value: int = value
        self._serialised_name: str = serialised_name

    @property
    def serialised_value(self) -> int:
        return self._serialised_value

    @property
    def serialised_name(self) -> str:
        return self._serialised_name

    @staticmethod
    def all() -> list[ShipUnlock]:
        return list(ShipUnlock)


class ExtraUnlock(Enum):
    orange_suit = 0
    green_suit = 1
    hazard_suit = 2
    pyjama_suit = 3
    purple_suit = 24

    @staticmethod
    def all() -> list[ExtraUnlock]:
        return list(ExtraUnlock)


class BestiaryEntry(Enum):
    snare_flea = 0
    bracken = 1
    thumper = 2
    eyeless_dog = 3
    hoarding_bug = 4
    hygroderes = 5
    slime = 5
    forest_keepers = 6
    giants = 6
    coil_head = 7
    spring_head = 7
    lasso_man = 8  # not implemented?
    earth_leviathan = 9
    sand_worm = 9
    jester = 10
    jack_in_the_box = 10
    spore_lizard = 11
    bunker_spider = 12
    spider = 12
    manticoil = 13
    circuit_bees = 14
    bees = 14
    roaming_locusts = 15
    locusts = 15
    baboon_hawk = 16
    nutcracker = 17

    @staticmethod
    def all() -> list[BestiaryEntry]:
        return list(BestiaryEntry)


class Item(Enum):
    binoculars = 0  # not yet implemented
    boom_box = 1
    cardboard_box = 2
    flashlight = 3
    jetpack = 4
    key = 5
    lockpick = 6
    handheld_monitor = 8  # not yet implemented
    pro_flashlight = 9
    shovel = 10
    flashbang = 11
    extension_ladder = 12
    tzp_inhalant = 13
    walkie_talkie = 14
    stun_gun = 15


class Scrap(Enum):
    apparatus = 7
    magic_7_ball = 16
    airhorn = 17
    bell = 18
    big_bolt = 19
    bottles = 20
    hairbrush = 21
    candy = 22
    cash_register = 23
    chemical_jug = 24
    clown_horn = 25
    large_axel = 26
    teeth = 27
    dustpan = 28
    egg_beater = 29
    v_type_engine = 30
    golden_cup = 31
    lamp = 32
    painting = 33
    plastic_fish = 34
    laser_pointer = 35
    gold_bar = 36
    hairdryer = 37
    magnifying_glass = 38
    tattered_metal_sheet = 39
    cookie_mold_pan = 40
    coffee_mug = 41
    perfume_bottle = 42
    old_phone = 43
    jar_of_pickles = 44
    pill_bottle = 45
    remote = 46
    ring = 47
    robot_toy = 48
    rubber_ducky = 49
    red_soda = 50
    steering_wheel = 51
    stop_sign = 52
    tea_kettle = 53
    toothpaste = 54
    toy_cube = 55
    bee_hive = 56
    radar_booster = 57
    yield_sign = 58
    shotgun = 59
    shotgun_shell = 60
    spray_paint = 61
    homemade_flashbang = 62
    gift_box = 63
    flask = 64
    tragedy = 65
    comedy = 66
    whoopie_cushion = 67


class Moon(Enum):
    experimentation = 0
    assurance = 1
    vow = 2
    company_building = 3
    march = 4
    rend = 5
    dine = 6
    offense = 7
    titan = 8



"""
The MIT License (MIT)

Copyright (c) 2023-present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TypedDict

from typing_extensions import Required

from .shared import *

__all__ = ("SaveFile",)


SaveFile = TypedDict(
    "SaveFile",
    {
        "CurrentPlanetID": Required[IntValue],
        "DeadlineTime": Required[IntValue],
        "EnemyScans": Required[ArrayIntValue],
        "FileGameVers": Required[IntValue],
        "GroupCredits": Required[IntValue],
        "ProfitQuota": Required[IntValue],
        "QuotaFulfilled": Required[IntValue],
        "QuotasPassed": Required[IntValue],
        "RandomSeed": Required[IntValue],
        "ShipUnlockStored_Cozy lights": Required[BoolValue],
        "ShipUnlockMoved_Cozy lights": BoolValue,
        "ShipUnlockPos_Cozy lights": VectorValue,
        "ShipUnlockRot_Cozy lights": VectorValue,
        "ShipUnlockStored_Teleporter": Required[BoolValue],
        "ShipUnlockMoved_Teleporter": BoolValue,
        "ShipUnlockPos_Teleporter": VectorValue,
        "ShipUnlockRot_Teleporter": VectorValue,
        "ShipUnlockStored_Television": Required[BoolValue],
        "ShipUnlockMoved_Television": BoolValue,
        "ShipUnlockPos_Television": VectorValue,
        "ShipUnlockRot_Television": VectorValue,
        "ShipUnlockStored_Cupboard": Required[BoolValue],
        "ShipUnlockMoved_Cupboard": BoolValue,
        "ShipUnlockPos_Cupboard": VectorValue,
        "ShipUnlockRot_Cupboard": VectorValue,
        "ShipUnlockStored_File Cabinet": Required[BoolValue],
        "ShipUnlockMoved_File Cabinet": BoolValue,
        "ShipUnlockPos_File Cabinet": VectorValue,
        "ShipUnlockRot_File Cabinet": VectorValue,
        "ShipUnlockStored_Toilet": Required[BoolValue],
        "ShipUnlockMoved_Toilet": BoolValue,
        "ShipUnlockPos_Toilet": VectorValue,
        "ShipUnlockRot_Toilet": VectorValue,
        "ShipUnlockStored_Shower": Required[BoolValue],
        "ShipUnlockMoved_Shower": BoolValue,
        "ShipUnlockPos_Shower": VectorValue,
        "ShipUnlockRot_Shower": VectorValue,
        "ShipUnlockStored_Light switch": Required[BoolValue],
        "ShipUnlockMoved_Light switch": BoolValue,
        "ShipUnlockPos_Light switch": VectorValue,
        "ShipUnlockRot_Light switch": VectorValue,
        "ShipUnlockStored_Record player": Required[BoolValue],
        "ShipUnlockMoved_Record player": BoolValue,
        "ShipUnlockPos_Record player": VectorValue,
        "ShipUnlockRot_Record player": VectorValue,
        "ShipUnlockStored_Table": Required[BoolValue],
        "ShipUnlockMoved_Table": BoolValue,
        "ShipUnlockPos_Table": VectorValue,
        "ShipUnlockRot_Table": VectorValue,
        "ShipUnlockStored_Romantic table": Required[BoolValue],
        "ShipUnlockMoved_Romantic table": BoolValue,
        "ShipUnlockPos_Romantic table": VectorValue,
        "ShipUnlockRot_Romantic table": VectorValue,
        "ShipUnlockStored_Bunkbeds": Required[BoolValue],
        "ShipUnlockMoved_Bunkbeds": BoolValue,
        "ShipUnlockPos_Bunkbeds": VectorValue,
        "ShipUnlockRot_Bunkbeds": VectorValue,
        "ShipUnlockStored_Terminal": Required[BoolValue],
        "ShipUnlockMoved_Terminal": BoolValue,
        "ShipUnlockPos_Terminal": VectorValue,
        "ShipUnlockRot_Terminal": VectorValue,
        "ShipUnlockStored_Signal translator": Required[BoolValue],
        "ShipUnlockMoved_Signal translator": BoolValue,
        "ShipUnlockPos_Signal translator": VectorValue,
        "ShipUnlockRot_Signal translator": VectorValue,
        "ShipUnlockStored_Loud horn": Required[BoolValue],
        "ShipUnlockMoved_Loud horn": BoolValue,
        "ShipUnlockPos_Loud horn": VectorValue,
        "ShipUnlockRot_Loud horn": VectorValue,
        "ShipUnlockStored_Inverse Teleporter": Required[BoolValue],
        "ShipUnlockMoved_Inverse Teleporter": BoolValue,
        "ShipUnlockPos_Inverse Teleporter": VectorValue,
        "ShipUnlockRot_Inverse Teleporter": VectorValue,
        "ShipUnlockStored_JackOLantern": Required[BoolValue],
        "ShipUnlockMoved_JackOLantern": BoolValue,
        "ShipUnlockPos_JackOLantern": VectorValue,
        "ShipUnlockRot_JackOLantern": VectorValue,
        "ShipUnlockStored_Welcome mat": Required[BoolValue],
        "ShipUnlockMoved_Welcome mat": BoolValue,
        "ShipUnlockPos_Welcome mat": VectorValue,
        "ShipUnlockRot_Welcome mat": VectorValue,
        "ShipUnlockStored_Goldfish": Required[BoolValue],
        "ShipUnlockMoved_Goldfish": BoolValue,
        "ShipUnlockPos_Goldfish": VectorValue,
        "ShipUnlockRot_Goldfish": VectorValue,
        "ShipUnlockStored_Plushie pajama man": Required[BoolValue],
        "ShipUnlockMoved_Plushie pajama man": BoolValue,
        "ShipUnlockPos_Plushie pajama man": VectorValue,
        "ShipUnlockRot_Plushie pajama man": VectorValue,
        "Stats_DaysSpent": Required[IntValue],
        "Stats_Deaths": Required[IntValue],
        "Stats_StepsTaken": Required[IntValue],
        "Stats_ValueCollected": Required[IntValue],
        "StoryLogs": Required[ArrayIntValue],
        "UnlockedShipObjects": Required[ArrayIntValue],
        "shipGrabbableItemIDs": ArrayIntValue,
        "shipGrabbableItemPos": ArrayVectorValue,
        "shipItemSaveData": ArrayIntValue,
        "shipScrapValues": ArrayIntValue,
    },
    total=False,
)
